#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "freeNN/feedforwardnetwork.h"
#include "csvparser.h"


#define TRAIN_POS_FILE "/home/tal/projects/HOG_dataset_orig/train_pos.csv"
#define TRAIN_NEG_FILE "/home/tal/projects/HOG_dataset_orig/train_neg.csv"
#define TEST_POS_FILE "/home/tal/projects/HOG_dataset_orig/test_pos.csv"
#define TEST_NEG_FILE "/home/tal/projects/HOG_dataset_orig/test_neg.csv"
#define NUM_INPUTS 6318


double **getInputsFromCsv(const char *filepath, int *num_samples_ref) {
	double **samples = NULL;
	*num_samples_ref = 0;
	CsvParser *csvparser = CsvParser_new(filepath, ",", 0);
	CsvRow *row;
	while ((row = CsvParser_getRow(csvparser)) ) {
		if ((*num_samples_ref) % 50 == 0) {
			printf("\rSamples loaded so far: %d", *num_samples_ref);
			fflush(stdout);
		}
		int num_fields = CsvParser_getNumFields(row);
		if (num_fields != NUM_INPUTS) {
			fprintf(stderr, "Error in CSV file: %s in line %d : Number of fields is %d instead of %d\n", filepath, *num_samples_ref, num_fields, NUM_INPUTS);
			exit(1);	
		}		
		samples = realloc(samples, ((*num_samples_ref) + 1) * sizeof(double*));
		samples[*num_samples_ref] = malloc(NUM_INPUTS * sizeof(double));		
		char **rowFields = CsvParser_getFields(row);
		int i;

		for (i = 0 ; i < num_fields ; i++) {
			samples[*num_samples_ref][i] = atof(rowFields[i]);
		}
		CsvParser_destroy_row(row);
		(*num_samples_ref)++;
	}
	CsvParser_destroy(csvparser);
	puts("");

	return samples;

}


double check_correctness(FeedForwardNetwork *ffn, double **pos, double **neg, int numPos, int numNeg) {
	int i;
	int num_neg_correct = 0;
	int num_pos_correct = 0;
	int num_incorrect = 0;
	for (i = 0 ; i < numNeg ; i++) {
		double *out = FeedForwardNetwork_activate(ffn, neg[i]);
		if (out[0] < 0.5) {
			num_neg_correct++;
		}
		free(out);
	}
	for (i = 0 ; i < numPos ; i++) {
		double *out = FeedForwardNetwork_activate(ffn, pos[i]);
		if (out[0] > 0.5) {
			num_pos_correct++;
		}
		free(out);
	}
	printf("Positive samples: %d of %d correct. Success rate: %f\n", num_pos_correct, numPos, (double)num_pos_correct / (double)numPos);
	printf("Negative samples: %d of %d correct. Success rate: %f\n", num_neg_correct, numNeg, (double)num_neg_correct / (double)numNeg);
	double crctNeg = (double)num_neg_correct / (double)numNeg;
	double crctPos = (double)num_pos_correct / (double)numPos;
	double TP_prop = (double)num_pos_correct / (double)numPos;
	double FP_prop = (double)(numNeg - num_neg_correct) / (double)numNeg;
	double precision = TP_prop / (TP_prop + FP_prop);
	double recall = (double)num_pos_correct / (double)numPos;

	double crct = sqrt(pow(precision, 2.0) + pow(recall, 2.0)) / sqrt(2.0);
	printf("Summary: %d of %d correct. Success rate: %f\n", num_neg_correct + num_pos_correct, numNeg + numPos, crct);
	printf("Success rate normalized: %f\n", crct);
	printf("Precision: %f , Recall: %f\n", precision, recall);

	return crct;
}

int check_binary_correctness(const double *expected, const double *actual) {
	if (expected[0] > 0.5 && actual[0] > 0.5) {
		return 1;
	}
	if (expected[0] < 0.5 && actual[0] < 0.2) {
		return 1;
	}
	return 0;
}


int main() {
	int num_train_pos;
	int num_train_neg;
	int num_test_pos;
	int num_test_neg;
	puts("Loading train pos...");
	double **train_pos = getInputsFromCsv(TRAIN_POS_FILE, &num_train_pos);
	puts("Loading train neg...");
	double **train_neg = getInputsFromCsv(TRAIN_NEG_FILE, &num_train_neg);
	puts("Loading test pos...");
	double **test_pos = getInputsFromCsv(TEST_POS_FILE, &num_test_pos);
	puts("Loading test neg...");
	double **test_neg = getInputsFromCsv(TEST_NEG_FILE, &num_test_neg);

	
	unsigned hidden[] = {50, 30};
	FeedForwardNetworkSettings *s = FeedForwardNetworkSettings_new_default();
	s->trainingRate_ = 0.01;
	s->outputLayerTrainingRate_ = 0.01;
	s->hiddenLayerType_ = TANH;
	s->outputLayerType_ = TANH;
	FeedForwardNetwork *ffn = FeedForwardNetwork_new(NUM_INPUTS, 1, 2, hidden, s);


	int num_epochs = 30000;
	int check_freq = 1;
	int learn_rate_freq = 200;
	int i, j;
	int k = 0;
	double pos_out[] = {1.0};
	double neg_out[] = {-1.0};

	double suc_rate = -1.0;
	int epochs_since_last_best = 0;

	char buffer[100] = "";
	int break_loop = 0;
	for (i = 0 ; i < num_epochs ; i++) {
		printf("\nepoch %d\n------------\n", i + 1);
		int pos_chg = 0;
		int neg_chg = 0;
		int yyy;

		for (j = 0 ; j < num_train_pos ; j++) {
			if (j % 50 == 0) {
				printf("\r%d%% completed", (int)(100 * (double)(j + 1) / (double)num_train_pos));
				fflush(stdout);
			}
			pos_chg += FeedForwardNetwork_train_fast(ffn, train_pos[j], pos_out, check_binary_correctness);
			for (yyy = (10 * j) ; yyy < (10 * j + 10) ; yyy++) {
				neg_chg += FeedForwardNetwork_train_fast(ffn, train_neg[yyy], neg_out, check_binary_correctness);
			}
		}
		puts("");
		printf("Changed %d pos samples and %d neg samples\n", pos_chg, neg_chg);
		if (i % check_freq == 0) {
			printf("Results against testing set:\n");
			double crct = check_correctness(ffn, test_pos, test_neg, num_test_pos, num_test_neg);
			printf("Last best was %f. Epochs since last best: %d. Num changed: %d\n", suc_rate, epochs_since_last_best, pos_chg + neg_chg);
			if (crct > suc_rate) {
				suc_rate = crct;
				epochs_since_last_best = 0;
				sprintf(buffer, "nets/net_%d", k);
				FeedForwardNetwork_saveToFile(ffn, buffer);
				printf("Wrote network to file: %s\n", buffer);
				k++;

			}
			epochs_since_last_best++;
	}

	return 0;
}

